from Interpreter.Repo import Repo
from Interpreter.Utils import Utils
from Models.Token import Token


class Lexer:

    def __init__(self, repo: Repo, utils: Utils) -> None:
        self.__repo = repo
        self.__utils = utils

        self.__ml_comment = False
        self.__toks = []

    def __add_tok(self, value: str, type: str) -> None:
        if value != '':
            if type == 'word':
                if value in self.__repo.keywords:
                    type = 'keyword'

                elif value in self.__repo.types:
                    type = 'type'

                elif value in self.__repo.bool_values:
                    type = 'bool'

                elif value == 'empty':
                    type = 'emptyType'

                else:
                    type = 'name'

            self.__toks.append(Token(type, value))

    def tokenize(self, line: str) -> list:
        self.__toks = []
        word = ''
        ol_comment = False
        is_string1 = False  # 'hello'
        is_string2 = False  # "hello"
        is_number = False
        is_expr_char = False

        for i in range(0, len(line)):
            c = line[i]
            # ignore comments and spaces
            if not ol_comment and not self.__ml_comment:

                # end of number
                if not c.isdigit() and is_number:
                    self.__add_tok(word, 'number')
                    word = ''
                    is_number = False

                # end of expression char
                if c not in self.__repo.expr_chars and is_expr_char:
                    self.__add_tok(word, 'expr_char')
                    word = ''
                    is_expr_char = False

                # comment filtering
                if c == '#' and not is_string1 and not is_string2:
                    ol_comment = True

                elif c == '/' and line[+1] == '/':
                    ol_comment = True

                elif c == '/' and line[+1] == '*':
                    self.__ml_comment = True
                    i += 2

                # begin of is_string1
                elif c == '\'' and not is_string1:
                    is_string1 = True
                    word = ''

                # end of is_string1
                elif c == '\'' and is_string1:
                    is_string1 = False
                    self.__add_tok(word, 'string')
                    word = ''

                # begin of is_string2
                elif c == '\"' and not is_string2:
                    is_string2 = True
                    word = ''

                # end of is_string2
                elif c == '\"' and is_string2:
                    is_string2 = False
                    self.__add_tok(word, 'string')
                    word = ''

                # format char
                elif c in self.__repo.format_chars:
                    self.__add_tok(word, 'word')
                    self.__add_tok(c, 'format_char')
                    word = ''

                # begin of number
                elif c.isdigit() and not is_number and word == '':
                    word += c
                    is_number = True

                # continue number
                elif c.isdigit() and is_number:
                    word += c

                # begin expression char
                elif c in self.__repo.expr_chars and not is_expr_char:
                    word += c
                    is_expr_char = True

                # continue expression char
                elif c in self.__repo.expr_chars and is_expr_char:
                    word += c

                # bool expression char
                elif c in self.__repo.expr_chars:
                    self.__add_tok(word, 'word')
                    self.__add_tok(c, 'bool_expr_char')
                    word = ''

                # end of word
                elif c == ' ' and not is_string1 and not is_string2 or c == '\n':
                    self.__add_tok(word, 'word')
                    word = ''

                else:
                    word += c

            if c == '\n' and ol_comment:
                ol_comment = False

            if c == '*' and line[i + 1] == '/':
                self.__ml_comment = False

        # self.__repo.output_tokens(self.__toks)
        return self.__toks
